#include <config.h>

#include <stdio.h>
#ifndef __USE_BSD
# define __USE_BSD	/* Needed for mkdtemp () */
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include "bmpx/uri.h"
#include "bmpx/vfs.h"
#include "bmpx/plugin.h"
#include "bmpx/plugin_interfaces.h"

#include "libtransport_cdda.h"

#ifdef HAVE_FTS_H
#  include <fts.h>
#endif

static VFSHandle *handle = NULL;

gboolean
transport_cdda_can_process (BmpPluginTransportCDDA *self, const gchar *uri)
{
    if (!g_ascii_strncasecmp("cdda",uri,4))
	return TRUE;

    return FALSE;
}

gboolean
transport_cdda_handle_set (BmpPluginTransportCDDA *self, VFSHandle *_handle)
{
    GError *error = NULL;

    if (_handle == NULL) 
	return FALSE;
    
    if (!_handle->valid)
	return FALSE;

    handle = _handle;

    return TRUE;
}

VFSHandle*
transport_cdda_handle_get (BmpPluginTransportCDDA *self)
{
    return handle;
}

gboolean
transport_cdda_handle_remove (BmpPluginTransportCDDA *self)
{
    handle = NULL;

    return TRUE;
}

gboolean
transport_cdda_handle_read (BmpPluginTransportCDDA *self)
{
    handle->uri_list = g_list_append (handle->uri_list, g_strdup(handle->uri));
    handle->need_container = FALSE;
    return TRUE;
}

static void
transport_cdda_interface_init (gpointer         g_iface,
                               gpointer         iface_data)
{
        BmpPluginTransportInterfaceClass *klass = (BmpPluginTransportInterfaceClass *)g_iface;

	klass->can_process    = transport_cdda_can_process;
	klass->handle_set     = transport_cdda_handle_set;
	klass->handle_get     = transport_cdda_handle_get;
	klass->handle_remove  = transport_cdda_handle_remove;
	klass->handle_read    = transport_cdda_handle_read;
}

GType plugin_transport_cdda_type = 0;

void 
plugin_transport_cdda_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginTransportCDDAClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginTransportCDDA),
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };

                static const GInterfaceInfo interface_transport_cdda_info = {
                        (GInterfaceInitFunc) transport_cdda_interface_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                plugin_transport_cdda_type = g_type_module_register_type (module,
					     G_TYPE_OBJECT,
					     "BmpPluginTransportCDDAType",
					     &object_info, 0);

                g_type_module_add_interface (module,
					     plugin_transport_cdda_type,
                                             BMP_TYPE_PLUGIN_TRANSPORT_IFACE,
                                             &interface_transport_cdda_info);
}
