/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Chroma widget set
 * (C) 2005 M. Derezynski <internalerror@gmail.com> 
 *
 */

#ifndef CHROMA_LIST_H
#define CHROMA_LIST_H

/*
 * Potentially, include other headers on which this header depends.
 */
#include <glib-object.h>
#include "bmp_ui.h"
#include "ui_util.h"
#include "chroma_vscrollbar.h"

#define CHROMA_TYPE_LIST		  (chroma_list_get_type ())
#define CHROMA_LIST(obj)		  (G_TYPE_CHECK_INSTANCE_CAST ((obj), CHROMA_TYPE_LIST, ChromaList))
#define CHROMA_LIST_CLASS(klass)	  (G_TYPE_CHECK_CLASS_CAST ((klass), CHROMA_TYPE_LIST, ChromaListClass))
#define CHROMA_IS_LIST(obj)	  (GTK_CHECK_TYPE ((obj), CHROMA_TYPE_LIST))
#define CHROMA_IS_LIST_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), CHROMA_TYPE_LIST))
#define CHROMA_LIST_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), CHROMA_TYPE_LIST, ChromaListClass))

typedef struct _ChromaList ChromaList;
typedef struct _ChromaListClass ChromaListClass;
typedef struct _ChromaListPrivate ChromaListPrivate;

struct _ChromaList {
	GObject parent;

        ChromaListPrivate *private;
};

struct _ChromaListClass {
	GObjectClass parent;

	void (*expose_request)    (ChromaList *list);
	void (*changed)           (ChromaList *list);
	void (*row_activated)     (ChromaList *list, gint _idx);
	void (*cursor_position)   (ChromaList *list, gint _idx);
};

typedef enum {
	CHROMA_LIST_ALIGN_LEFT,
	CHROMA_LIST_ALIGN_CENTER
} ChromaListAlignMode;

typedef enum {
	CHROMA_LIST_DROP_ROWS,
	CHROMA_LIST_DROP_WHOLE,
} ChromaListDropMode;

/* used by CHROMA_TYPE_LIST */
GType chroma_list_get_type (void);

GObject*
chroma_list_new (GtkWidget *parent, gint position_x, gint position_y, gint width, gint height); 

void
chroma_list_draw (ChromaList *self, GtkWidget *widget, cairo_t *cr, GdkVisibilityState state);

gboolean
chroma_list_get_changed (ChromaList *self);

void
chroma_list_get_size (ChromaList *self, GtkAllocation *allocation);
 
void
chroma_list_emit_expose_request (ChromaList *self);

void
chroma_list_move_resize(ChromaList * self, gint pos_x, gint pos_y, gint w, gint h);

void
chroma_list_mark_dirty (ChromaList *self);
                      
void
chroma_list_list_changed (ChromaList *self);

void
chroma_list_emit_changed (ChromaList *self);
	
gint
chroma_list_get_drop_position (ChromaList *self);
 
gboolean
chroma_list_dispatch_drag_leave (ChromaList *self);
 
gboolean
chroma_list_dispatch_drag_motion (ChromaList *self, gint x, gint y); 

gboolean
chroma_list_dispatch_button_press (ChromaList *self, GdkEventButton *event); 

gboolean
chroma_list_dispatch_button_release (ChromaList *self, GdkEventButton *event);

gboolean
chroma_list_dispatch_key_press(ChromaList *self, GdkEventKey *event);
 
gboolean
chroma_list_dispatch_motion (ChromaList *self, GdkEventMotion *event); 

gboolean
chroma_list_inside(ChromaList *self, gint x,  gint y);
	
gint
chroma_list_get_n_item_display (ChromaList *self);

gint
chroma_list_get_n_items (ChromaList *self);
 
void
chroma_list_set_list (ChromaList *self, const BmpListType *list);

const BmpListType*
chroma_list_get_list (ChromaList *self);
	
gint
chroma_list_get_last_selected (ChromaList *self);

void
chroma_list_set_last_selected (ChromaList *self, gint _idx);
	
GList*
chroma_list_get_selected_paths (ChromaList *self);

gint
chroma_list_get_row_height (ChromaList *self);
 
void
chroma_list_select_nth (ChromaList *self, gint _idx, gboolean selected);

void
chroma_list_select_all (ChromaList *self);

void
chroma_list_unselect_all (ChromaList *self);

void
chroma_list_invert_selection (ChromaList *self);

void
chroma_list_select_name (ChromaList *self, const gchar *name);

gchar*
chroma_list_get_selected_name (ChromaList *self);

/* Parameters */
void
chroma_list_set_font_information (ChromaList *self, BmpFontInformation *font_information);

void
chroma_list_set_adjustment (ChromaList *self, ChromaVScrollbar *adjustment);
 
void
chroma_list_set_use_numbers (ChromaList *self, gboolean use_numbers); 

void
chroma_list_set_columns (ChromaList *self, gint columns);

void
chroma_list_set_selection_mode (ChromaList *self, GtkSelectionMode selection_mode);
	
void
chroma_list_set_align (ChromaList *self, ChromaListAlignMode align);

void
chroma_list_set_drop_mode (ChromaList *self, ChromaListDropMode mode);

void
chroma_list_set_change_on_select (ChromaList *self, gboolean change_on_select);
/* END Parameters */

#endif /* CHROMA_LIST_H */



