#include <string.h>
#include <stdlib.h>

#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>
#include <X11/Xlib.h>

#include <main.h>
#include <bmp_ui.h>
#include <dnd.h>

#include <ui_main.h>
#include <ui_playlist.h>
#include <ui_util.h>

/* bmp_dock.c/bmp_dock.h (C) 1998-2003 XMMS.org, (C) 2004-2005 BMP DT */
#include <bmp_dock.h>
#include "bmp_window.h"
#include "bmp_button.h"
#include <chroma_vscrollbar.h>
#include <chroma_list.h>
#include <chroma_button.h>
#include <g_object_get.h>
#include "bmpx/util.h"
#include <gc.h>

#include "bmpx/uri.h"
#include "bmpx/playlist.h"

typedef struct {

	gchar * name;
	guint x_origin;
	guint y_origin;
	guint width;
	guint height;
	GdkPixbuf *pixbuf;

} BmpUIComponent;

#define BMP_UIC_PL_INFORMATION  "bmp_uic_pl_information"

#define BMP_UIC_PL_TITLE        "bmp_uic_pl_title"
#define BMP_UIC_PL_TITLE_NOFOCUS     "bmp_uic_pl_title_nf"

#define BMP_UIC_PL_PAD_TOP      "bmp_uic_pl_pad_top"
#define BMP_UIC_PL_PAD_TOP_NOFOCUS   "bmp_uic_pl_pad_top_nf"

#define BMP_UIC_PL_PAD_BOTTOM   "bmp_uic_pl_pad_bottom"

#define BMP_UIC_PL_CORNER_UL    "bmp_uic_pl_corner_ul"
#define BMP_UIC_PL_CORNER_UL_NOFOCUS "bmp_uic_pl_corner_ul_nf"

#define BMP_UIC_PL_CORNER_UR    "bmp_uic_pl_corner_ur"
#define BMP_UIC_PL_CORNER_UR_NOFOCUS "bmp_uic_pl_corner_ur_nf"

#define BMP_UIC_PL_EDGE_LEFT    "bmp_uic_pl_edge_left"
#define BMP_UIC_PL_EDGE_RIGHT   "bmp_uic_pl_edge_right"

#define BMP_UIC_PL_CORNER_LL    "bmp_uic_pl_corner_ll"
#define BMP_UIC_PL_CORNER_LR    "bmp_uic_pl_corner_lr"

static BmpUIComponent bmp_playlist_component_data[] = {

  {BMP_UIC_PL_TITLE,         26,  0, 100, 20 },
  {BMP_UIC_PL_TITLE_NOFOCUS,      26, 21, 100, 20 },

  {BMP_UIC_PL_PAD_TOP,      127,  0,  25, 20 },
  {BMP_UIC_PL_PAD_TOP_NOFOCUS,   127, 21,  25, 20 },

  {BMP_UIC_PL_PAD_BOTTOM,   179,  0,  25, 38 },

  {BMP_UIC_PL_CORNER_UL,      0,  0,  25, 20 },
  {BMP_UIC_PL_CORNER_UL_NOFOCUS,   0, 21,  25, 20 },

  {BMP_UIC_PL_CORNER_UR,    153,  0,  25, 20 },
  {BMP_UIC_PL_CORNER_UR_NOFOCUS, 153, 21,  25, 20 },

  {BMP_UIC_PL_EDGE_LEFT,      0, 42,  12, 29 },
  {BMP_UIC_PL_EDGE_RIGHT,    32, 42,  19, 29 },

  {BMP_UIC_PL_CORNER_LL,      0, 72, 125, 38 },
  {BMP_UIC_PL_CORNER_LR,    126, 72, 150, 38 },

  {BMP_UIC_PL_INFORMATION,  205,  0,  75, 32 }

};

#define VSCROLLBAR_DEFAULT_WIDTH 12
#define VSEP_MIN_WIDTH 192
#define VSEP_SNAP_WIDTH 24

#define PLEDIT_INNER_XMARGIN 21
#define PLEDIT_INNER_YMARGIN 58

#define PLEDIT_ROOT_X 12
#define PLEDIT_ROOT_Y 20

#define PLEDIT_INNERPAD 2

#define PLEDIT_HEADER_TEXT_PLAYLISTS "Playlists"
#define PLEDIT_HEADER_TEXT_LIBRARY   "Library"
#define PLEDIT_HEADER_TEXT_TRACKLIST "Track List"

#define REGION_L(x1,x2,y1,y2,w,h)                   \
    (event->x >= (x1) && event->x < (x2) &&     \
     event->y >= h - (y1) &&  \
     event->y < h - (y2))

#define REGION_R(x1,x2,y1,y2,w,h) \
    ((event->x >= (w-x1)) && \
     (event->x < (w-x2)) && \
     (event->y >= (h-y1)) && \
     (event->y < (h-y2)))

static GtkActionEntry bmp_actions_PL[] = {

  { BMP_ACTION_PLAYLIST_ADD,
	GTK_STOCK_NEW,
	N_("New Playlist"),NULL,
	NULL, G_CALLBACK (bmp_ui_callback_tracklist_new)},
  { BMP_ACTION_PLAYLIST_DELETE,
	GTK_STOCK_DELETE,
	N_("Remove Playlist"),"<shift>Delete",
	NULL, G_CALLBACK (bmp_ui_callback_tracklist_remove)},
  { BMP_ACTION_PLAYLIST_DELETE_ALL,
	GTK_STOCK_DELETE,
	N_("Remove All Playlists"),"<control><shift>Delete",
	NULL, G_CALLBACK (bmp_ui_callback_tracklists_remove_all)},
  { BMP_ACTION_PLAYLIST_RENAME,
	GTK_STOCK_EDIT,
	N_("Rename Playlist"),"F2",
	NULL, G_CALLBACK (bmp_ui_callback_tracklist_rename)},
  { BMP_ACTION_PLAYLIST_SET_FILENAME,
	GTK_STOCK_SAVE,
	N_("Set Playlist Filename"),NULL,
	NULL, G_CALLBACK (bmp_ui_callback_set_playlist_filename)},

  { BMP_ACTION_TRACKLIST_FILEINFO,
	GTK_STOCK_INFO,    N_("File Information"),      "I",
	NULL, G_CALLBACK (bmp_ui_callback_fileinfo)},

  { BMP_ACTION_TRACKLIST_ADD_URL,
	GTK_STOCK_NETWORK,   N_("Add Internet Stream..."),  "<shift>U",
	NULL, G_CALLBACK (bmp_ui_callback_add_url)},
  { BMP_ACTION_TRACKLIST_ADD_CD,
	GTK_STOCK_CDROM,   N_("Add CD..."),             "<shift>C",
	NULL, G_CALLBACK (bmp_ui_callback_add_cd)},
  { BMP_ACTION_TRACKLIST_ADD,
	GTK_STOCK_ADD,    N_("Add Files..."),           "F",
	NULL, G_CALLBACK (bmp_ui_callback_add_files)},
  { BMP_ACTION_TRACKLIST_ADD_LIBRARY,
	GTK_STOCK_ADD,    N_("Add From Library..."),   "<control>F",
	NULL, G_CALLBACK (bmp_ui_callback_add_files_library)},
  { BMP_ACTION_TRACKLIST_ADD_TO_NEW,
	GTK_STOCK_ADD,    N_("Add Files to New..."),  "<shift>F",
	NULL, G_CALLBACK (NULL)},
  { BMP_ACTION_TRACKLIST_DEL,
	GTK_STOCK_DELETE, N_("Remove Selected"),       "Delete",
	NULL, G_CALLBACK (bmp_ui_callback_del_files)},
  { BMP_ACTION_TRACKLIST_CROP,
	GTK_STOCK_CUT,    N_("Remove Unselected"),     "<ctrl>X",
	NULL, G_CALLBACK (bmp_ui_callback_crop_files)},
  { BMP_ACTION_TRACKLIST_DEL_ALL,
	GTK_STOCK_CLEAR,  N_("Remove All"),            NULL,
	NULL, G_CALLBACK (bmp_ui_callback_del_files_all)},

  { BMP_ACTION_CLEAR_PLAYBACK_HISTORY,
	GTK_STOCK_UNDO,  N_("Clear Playback History"), "<shift>e",
	NULL, G_CALLBACK (bmp_ui_callback_clear_playback_history)},

  { BMP_ACTION_TRACKLIST_SELECT_ALL,
	NULL,N_("Select All"),            "<control>a",
	NULL, G_CALLBACK (bmp_ui_callback_select_all_files)},
  { BMP_ACTION_TRACKLIST_SELECT_NONE,
	NULL,N_("Select None"),           "<control><shift>a",
	NULL, G_CALLBACK (bmp_ui_callback_unselect_all_files)},
  { BMP_ACTION_TRACKLIST_SELECT_INVERT,
	NULL,N_("Invert Selection"),      NULL,
	NULL, G_CALLBACK (bmp_ui_callback_invert_selection)},


  { BMP_ACTION_SORT_VISIBLE,
	NULL,N_("Sort by Visible Title"),      "<shift>s",
	NULL, G_CALLBACK (bmp_ui_callback_sort_current_list)},
  { BMP_ACTION_SORT_ARTIST,
	NULL,N_("Sort by Artist"),             NULL,
	NULL, G_CALLBACK (bmp_ui_callback_sort_current_list_artist)},
  { BMP_ACTION_SORT_ALBUM,
	NULL,N_("Sort by Album"),              NULL,
	NULL, G_CALLBACK (bmp_ui_callback_sort_current_list_album)},
  { BMP_ACTION_SORT_TITLE,
	NULL,N_("Sort by Title"),              NULL,
	NULL, G_CALLBACK (bmp_ui_callback_sort_current_list_title)},
  { BMP_ACTION_SORT_GENRE,
	NULL,N_("Sort by Genre"),              NULL,
	NULL, G_CALLBACK (bmp_ui_callback_sort_current_list_genre)},
  { BMP_ACTION_SORT_TRACK,
	NULL,N_("Sort by Tracknumber"),        NULL,
	NULL, G_CALLBACK (bmp_ui_callback_sort_current_list_track)},
  { BMP_ACTION_SORT_LOCATION,
	NULL,N_("Sort by Location"),           NULL,
	NULL, G_CALLBACK (bmp_ui_callback_sort_current_list_location)}


};

BmpFontInformation* bmp_fonts[BMP_N_FONTS];

enum {
    BMP_STAT_VSEP_WIDTH,

    BMP_N_UI_VALS
} BmpPLUIVals;

enum {
    BMP_STAT_VSEP_INSIDE,
    BMP_STAT_VSEP_DRAG,
    BMP_STAT_VSEP_SNAPPED,

    BMP_N_BOOL_STATS
} BmpPLBoolStats;

/* FIXME: Replace trough properties */
static void
window_playlist_create(BmpWindowPlaylist *self);

static void
window_playlist_dispose (GObject *gobject);

static void
resize_widgets (BmpWindowPlaylist *self, gint w, gint h);

static void
tracklist_adapt_actions (BmpWindowPlaylist *self, ChromaList *list);

static void
set_vsep_width (BmpWindowPlaylist *self, gint width);

static void
on_window_playlist_scroll(GtkWidget *widget,GdkEventScroll *scroll_event,gpointer data);

/* GObject stuff */
G_DEFINE_TYPE (BmpWindowPlaylist, bmp_window_playlist, G_TYPE_OBJECT)

struct _BmpWindowPlaylistPrivate {

    gboolean dispose_has_run;

    GHashTable *ui_components;
    GPtrArray  *array_buttons;

    GdkVisibilityState visibility;

    gint     statvals_int[BMP_N_UI_VALS];
    gboolean statvals_bool[BMP_N_BOOL_STATS];
};

void
destroy_value(gpointer data) {

  BmpUIComponent *uic;
  uic = (BmpUIComponent *) data;
  g_object_unref(uic->pixbuf);
  g_free(uic);

}

static void
update_view (BmpWindowPlaylist *self, gpointer data)
{
	ChromaList *list;

	list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
	chroma_list_list_changed (list);

	list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget);
	chroma_list_list_changed (list);

	gtk_widget_queue_draw (self->window);
}

#if 0
static bmp_button_normal bmp_buttons_window_playlist[] =
{
    /*
    {skin file enum, xcoord normal, ycoord normal,
                     xcoord pressed, ycoord pressed,
		     width, height,
		     xpos on skin, ypos on skin,
		     clickable}
     */

    /* Add button */
    /* {BMP_ACTION_ADD, SKIN_COMPONENT_PLEDIT, 0, 149, 23, 149,  22,  18, 0, 0, BMP_WINDOW_PLEDIT }, */
}
gint nbuttons_window_playlist = G_N_ELEMENTS(bmp_buttons_window_playlist);
#endif

static BmpUIComponent*
playlist_get_component(BmpWindowPlaylist *self, gchar * key) {

	gpointer p;

	p = g_hash_table_lookup(self->private->ui_components, key);

	return (BmpUIComponent *) p;

}

static void
window_playlist_set_geometry_hints(BmpWindowPlaylist *self)
{
    BmpWindow *window = BMP_WINDOW(self->window);
    GdkGeometry geometry;
    GdkWindowHints mask;

#if 0
    geometry.min_width = 25*20; 
    geometry.max_width = 0xFFFF;
    geometry.base_width = 25*20; 
#endif

    geometry.width_inc = 25;
    geometry.height_inc = 29;

    /* make sure the min width/height are multiple values of the resize increments */
    geometry.min_width = ((int) (BMP_WINDOW_MAIN_WIDTH * 2 / geometry.width_inc)) * geometry.width_inc;
    geometry.max_width = 0xFFFF;
    geometry.base_width = 0;

    /* make sure the min width/height are multiple values of the resize increments */
    geometry.min_height = ((int) (BMP_WINDOW_MAIN_HEIGHT * 2 / geometry.height_inc)) * geometry.height_inc;
    geometry.max_height = 0xFFFF;
    geometry.base_height = 0;

#if 0
    geometry.min_height = 29*10; 
    geometry.max_height = 0xFFFF;
    geometry.base_height = 29*10; 
#endif

    mask = GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE | GDK_HINT_RESIZE_INC | GDK_HINT_BASE_SIZE;

    gtk_window_set_geometry_hints  (GTK_WINDOW(window),
                                    GTK_WIDGET(window),
                                    &geometry,
				    mask);

}

static void
draw_header (BmpWindowPlaylist *self, 
	     GtkWidget *widget, 
	     cairo_t *cr, 
	     gint x,
             gint y, 
             gint width, 
             gint height, 
             const gchar* text, 
             const gchar* info_text)
{

#define HEADER_PAD 8

    PangoAttrList *attr_list;
    PangoAttribute *attr;
    PangoLayout *layout;
    PangoRectangle rect1, rect2;
    gchar *aux;
    gint l_width1 = 0, l_width2 = 0, aux_int = 0;
    double r,g,b,a;

    g_return_if_fail (GTK_IS_WIDGET(widget));

    gc_mark(ngc);

    bmp_cairo_draw_rounded_rectangle (cr, x, y, width, height, bmp_colors_get_color (BMP_SKINCOLOR_PL_FG_CURRENT), 0.2, 1.0, TRUE, 0.6);
    bmp_cairo_draw_rounded_rectangle (cr, x, y, width, height, bmp_colors_get_color (BMP_SKINCOLOR_PL_FG_CURRENT), 1.0,  .4, FALSE, 0.6);

    gdk_color_to_rgba (bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT), &r, &g, &b, &a);

    attr_list = pango_attr_list_new();
    attr = pango_attr_foreground_new (r * 65535, g * 65535, b * 65535);
    pango_attr_list_insert (attr_list, attr);

    layout = pango_cairo_create_layout (cr);
    pango_layout_set_attributes (layout, attr_list);

    if (info_text != NULL)
      {
	    aux = g_strconcat ("<span size=\"x-small\" weight=\"bold\">",   
		gc_collect(ngc, g_markup_escape_text (info_text, strlen(info_text))), "</span>", 
		NULL);

	    pango_layout_set_markup (layout, aux, strlen(aux));
	    pango_layout_set_font_description (layout,
		    bmp_fonts[BMP_FONT_LARGE]->font_description);
	    g_free (aux);

	    pango_layout_get_pixel_extents (layout, &rect1, &rect2);
	    l_width1 = rect1.width;
	    l_width1 += HEADER_PAD/2;

	    cairo_move_to (cr, (x+width) - rect1.width - (HEADER_PAD*1.5) - aux_int, y + (rect1.height/2) - 1);
	    cairo_set_source_rgb (cr, r,g,b);
	    pango_cairo_show_layout (cr, layout);
      }

    l_width2 = l_width1*2;

    if (text != NULL)
      {
        aux = g_strconcat ("<b>", gc_collect(ngc, g_markup_escape_text (text, strlen(text))), "</b>", NULL);
	pango_layout_set_markup (layout, aux, strlen(aux));
        pango_layout_set_font_description (layout, bmp_fonts[BMP_FONT_LARGE]->font_description);
	pango_layout_set_width (layout, (width-(l_width2))*PANGO_SCALE);
        pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
	pango_layout_set_alignment (layout, PANGO_ALIGN_LEFT);

	pango_layout_get_pixel_extents (layout, &rect1, &rect2);

	cairo_move_to (cr, x+6, y + ((height - rect2.height)/2));
	cairo_set_source_rgb (cr, r,g,b);
	pango_cairo_show_layout (cr, layout);

	g_free (aux);
        g_object_unref(layout);
      }

    gc_free (ngc);
}

static gboolean
on_window_playlist_visibility (GtkWidget *widget, GdkEventVisibility *event, BmpWindowPlaylist *self)
{
	self->private->visibility = event->state;

	return FALSE;
}

static gboolean
on_window_playlist_delete_event (GtkWidget *widget, GdkEvent *event, BmpWindowPlaylist *self)
{
    gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(gtk_action_group_get_action (bmp_toggle_actions, BMP_TOGGLE_ACTION_PL)), FALSE);

    return TRUE;
}
 
static gboolean
on_window_playlist_state_event (GtkWidget *widget, GdkEventWindowState *event, BmpWindowPlaylist *self)
{
    BmpWindowMain *window_main;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if ( ((event->changed_mask & GDK_WINDOW_STATE_ICONIFIED) == GDK_WINDOW_STATE_ICONIFIED)	&&
	     ((event->new_window_state & GDK_WINDOW_STATE_ICONIFIED) == 0))
    {
	window_main = RM_GET_ITEM("windows", "window-main", NOLOCK);
	gtk_window_deiconify (GTK_WINDOW(window_main->window));
	gtk_window_present (GTK_WINDOW(window_main->window));
    }

    return FALSE;
}


static gboolean
on_window_playlist_focus_in (GtkWidget *widget, GdkEventFocus *event, BmpWindowPlaylist *self)
{
    BmpWindowMain *window_main;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    window_main = RM_GET_ITEM("windows", "window-main", NOLOCK);
    gdk_window_raise (window_main->window->window);
    gtk_widget_queue_draw (self->window);

    return FALSE;
}

static gboolean
on_window_playlist_focus_out (GtkWidget *widget, GdkEventFocus *event, BmpWindowPlaylist *self)
{
	g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

	return FALSE;
}

/* Drawing */
static void
draw_list_playlists (BmpWindowPlaylist *self, ChromaList *list_playlists, ChromaVScrollbar *vs_playlists, cairo_t *cr, gint w_width, gint w_height, gint vsep_width, gboolean draw_list)
{
    GtkWidget *widget = BMP_WINDOW(self->window)->canvas;
    gint aux0, aux1;

    aux0 = g_object_get_int (list_playlists, "n_items");
    aux1 = chroma_list_get_n_item_display (list_playlists);

    if (draw_list) chroma_list_draw (list_playlists, widget, cr, self->private->visibility);
    if (aux0 > aux1) chroma_vscrollbar_draw (vs_playlists, widget->window, cr);

}

static void
draw_list_tracklist (BmpWindowPlaylist *self, ChromaList *list_tracklist, ChromaVScrollbar *vs_tracklist, cairo_t *cr, gint w_width, gint w_height, gint vsep_width, gboolean draw_list)
{
    GtkWidget *widget = BMP_WINDOW(self->window)->canvas;
    gint aux0, aux1;

    aux0 = g_object_get_int (list_tracklist, "n_items");
    aux1 = chroma_list_get_n_item_display (list_tracklist);

    if (draw_list) chroma_list_draw (list_tracklist, widget, cr, self->private->visibility);
    if (aux0 > aux1) chroma_vscrollbar_draw (vs_tracklist, widget->window, cr);
}

static gboolean
on_window_playlist_canvas_expose (GtkWidget *widget, GdkEventExpose *event, BmpWindowPlaylist *self)
{
  GdkGC *gc, *gc_save;
  GdkRectangle *c_rect;
  BmpUIComponent *uic;
  ChromaList *list_tracklist, *list_playlists;
  ChromaVScrollbar *vs_tracklist, *vs_playlists;
  cairo_t *cr;
  guint w_width, w_height, pad_ctr = 0, i = 0, vsep_width = 0;
  gint ax, ay, aw, ah, cx, cy, cw, ch, lx, ly, lw, lh;
  gboolean focus = gtk_window_has_toplevel_focus (GTK_WINDOW (self->window));
  double r,g,b,a;
  const gchar *aux_str0;
  gchar *aux_str1;

  w_width = widget->allocation.width;
  w_height = widget->allocation.height;

  list_tracklist = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
  list_playlists = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget);

  vs_tracklist = CHROMA_VSCROLLBAR(chroma_widgets[BMP_CHROMA_VSCROLLBAR_TRACKLIST]->widget);
  vs_playlists = CHROMA_VSCROLLBAR(chroma_widgets[BMP_CHROMA_VSCROLLBAR_PLAYLISTS]->widget);

  ax = event->area.x;
  ay = event->area.y;
  aw = event->area.width;
  ah = event->area.height;

  cx = PLEDIT_ROOT_X;
  cy = PLEDIT_ROOT_Y;
  cw = event->area.width + event->area.x;
  /*FIXME: hack */
  ch = event->area.height + event->area.y - 2;

  cr = gdk_cairo_create (widget->window);

  gc_save = gdk_gc_new (widget->window);
  gdk_gc_copy (gc_save, widget->style->fg_gc[GTK_STATE_NORMAL]);
  gc = widget->style->fg_gc[GTK_STATE_NORMAL];

  pad_ctr =  (w_width - 150)/25;

  if (self->private->statvals_int[BMP_STAT_VSEP_WIDTH] > (w_width/2))
      self->private->statvals_int[BMP_STAT_VSEP_WIDTH] = w_width / 2;

  vsep_width = self->private->statvals_int[BMP_STAT_VSEP_WIDTH];

  if (focus)
	  uic = playlist_get_component(self, BMP_UIC_PL_CORNER_UL);
  else
	  uic = playlist_get_component(self, BMP_UIC_PL_CORNER_UL_NOFOCUS);
  bmp_draw_pixbuf_simple(cr, uic->pixbuf, 0, 0);

  if (focus)
	  uic = playlist_get_component(self, BMP_UIC_PL_CORNER_UR);
  else
	  uic = playlist_get_component(self, BMP_UIC_PL_CORNER_UR_NOFOCUS);
  bmp_draw_pixbuf_simple(cr, uic->pixbuf, (w_width-uic->width), 0);

  if (focus)
  	uic = playlist_get_component(self, BMP_UIC_PL_TITLE);
  else
  	uic = playlist_get_component(self, BMP_UIC_PL_TITLE_NOFOCUS);
  bmp_draw_pixbuf_simple(cr, uic->pixbuf, (w_width/2)-(50), 0);

  uic = playlist_get_component(self, BMP_UIC_PL_CORNER_LL);
  bmp_draw_pixbuf_simple(cr, uic->pixbuf, 0, (w_height-uic->height));

  uic = playlist_get_component(self, BMP_UIC_PL_CORNER_LR);
  bmp_draw_pixbuf_simple(cr, uic->pixbuf, (w_width-uic->width), (w_height-uic->height));

  if (focus)
	  uic = playlist_get_component(self, BMP_UIC_PL_PAD_TOP);
  else
	  uic = playlist_get_component(self, BMP_UIC_PL_PAD_TOP_NOFOCUS);

    for (i = 0; i < pad_ctr / 2; i++) {

        /* Left of title */
	bmp_draw_pixbuf_simple(cr, uic->pixbuf, 25 + i * 25, 0);

        /* Right of title */
	bmp_draw_pixbuf_simple(cr, uic->pixbuf, (w_width+100)/2+i*25, 0);
    }

    if ( pad_ctr & 1) {

        /* Odd tile count, so one remaining to draw. Here we split
         * it into two and draw half on either side of the title */

	    gdk_draw_pixbuf(widget->window,
                        NULL,
                        uic->pixbuf,
			0,
                        0,
                        (w_width/2)+((pad_ctr/2)*25) +50,
                        0,
                        13,
                        -1,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);

	    gdk_draw_pixbuf(widget->window,
                        NULL,
                        uic->pixbuf,
			0,
                        0,
			((pad_ctr / 2) * 25) + 25,
                        0,
                        13,
                        -1,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);
    }

    /* Frame sides */
    for (i = 0; i < (w_height - (20 + 38)) / 29; i++) {

            uic = playlist_get_component(self, BMP_UIC_PL_EDGE_LEFT);
            gdk_draw_pixbuf(widget->window,
                        NULL,
                        uic->pixbuf,
                        0,
                        0,
			0,
                        20 + i *29,
                        -1,
                        -1,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);

            uic = playlist_get_component(self, BMP_UIC_PL_EDGE_RIGHT);
            gdk_draw_pixbuf(widget->window,
                        NULL,
                        uic->pixbuf,
                        0,
                        0,
			w_width-19,
                        20 + i *29,
                        -1,
                        -1,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);

    }

    pad_ctr =  (w_width - 275) / 25;
    uic = playlist_get_component(self, BMP_UIC_PL_PAD_BOTTOM);
    for (i = 0; i < pad_ctr+4; i++)
	bmp_draw_pixbuf_simple(cr, uic->pixbuf, (125+i*25), (w_height-uic->height));
    bmp_draw_pixbuf_simple(cr, uic->pixbuf, (125+(pad_ctr+3)*25) + 6, (w_height-uic->height));

    c_rect = g_new0(GdkRectangle, 1);
    c_rect->x = 0;
    c_rect->y = 0;
    c_rect->width = w_width - 15;
    c_rect->height = w_height - PLEDIT_INNER_YMARGIN;

    gdk_gc_set_clip_origin(gc, PLEDIT_ROOT_X, PLEDIT_ROOT_Y);
    gdk_gc_set_clip_rectangle(gc, c_rect);

    /* Draw main vertical separator */
    lx = w_width - vsep_width;
    ly = PLEDIT_ROOT_Y - PLEDIT_INNERPAD + 8 + 1;
    lw = 2;
    lh = w_height - PLEDIT_INNER_YMARGIN - PLEDIT_ROOT_Y;

    gdk_color_to_rgba (bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT), &r, &g, &b, &a);
    cairo_new_path(cr);
    cairo_set_line_width(cr, 1);
    cairo_rectangle(cr, lx, ly, lw, lh);
    cairo_set_source_rgba(cr, r, g, b, 0.5);
    cairo_set_line_cap (cr, CAIRO_LINE_CAP_ROUND);
    cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
    cairo_set_antialias (cr, CAIRO_ANTIALIAS_DEFAULT);
    cairo_fill(cr);

    /* "Track List" header */
    aux_str0 = bmp_playlist_get_nth_tracklist_name (bmp_playlist, g_object_get_int (bmp_system_control, "current_tracklist"));
    aux_str1 = (chroma_list_get_n_items(list_tracklist)>0) ? g_strdup_printf("%d %s", chroma_list_get_n_items(list_tracklist), (chroma_list_get_n_items(list_tracklist) > 1) ? "Tracks" : "Track") : NULL;
    draw_header (self, widget, cr,
			    PLEDIT_ROOT_X + PLEDIT_INNERPAD*2,
			    PLEDIT_ROOT_Y + PLEDIT_INNERPAD+4,
			    w_width - vsep_width - PLEDIT_ROOT_X - 8 - PLEDIT_INNERPAD,
			    bmp_fonts[BMP_FONT_LARGE]->font_height + 2,
			    aux_str0, aux_str1);
    g_free (aux_str1);

    /* Chroma List "Playlists" */
    if ((!self->private->statvals_bool[BMP_STAT_VSEP_SNAPPED]))
	draw_list_playlists (self, list_playlists, vs_playlists, cr, w_width, w_height, vsep_width, TRUE);

    /* Chroma List "Track List" */
    draw_list_tracklist (self, list_tracklist, vs_tracklist, cr, w_width, w_height, vsep_width, TRUE);

    gdk_gc_set_clip_origin(gc, 0, 0);
    gdk_gc_set_clip_rectangle(gc, NULL);
    gdk_gc_copy (widget->style->fg_gc[GTK_STATE_NORMAL], gc_save);
    cairo_destroy (cr);

    return FALSE;
}

/* Action adapters */
static void
playlists_adapt_actions (BmpWindowPlaylist *self, gboolean active)
{
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PLAYLIST_DELETE_ALL), active);
}

static void
playlists_adapt_actions_changed (BmpWindowPlaylist *self, gboolean active)
{
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_ADD_LIBRARY), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_ADD), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_ADD_CD), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_ADD_URL), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PLAYLIST_DELETE), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PLAYLIST_RENAME), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PLAYLIST_SET_FILENAME), active);
}

static void
tracklist_adapt_actions_has_sel (BmpWindowPlaylist *self, gboolean active)
{
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_DEL), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_CROP), active);
}

static void
tracklist_adapt_actions_has_items (BmpWindowPlaylist *self, gboolean active)
{
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_DEL_ALL), active);
}

static void
tracklist_adapt_actions_sel (BmpWindowPlaylist *self, gboolean active)
{
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_SELECT_INVERT), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_SELECT_NONE), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_SELECT_ALL), active);
}

static void
tracklist_adapt_actions_sort (BmpWindowPlaylist *self, gboolean active)
{
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_SORT_LOCATION), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_SORT_TRACK), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_SORT_VISIBLE), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_SORT_ARTIST), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_SORT_ALBUM), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_SORT_TITLE), active);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_SORT_GENRE), active);
}

static void
on_bmp_font_changed (BmpWindowPlaylist *self)
{
    chroma_list_set_font_information (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget), bmp_fonts[BMP_FONT_NORMAL]);
    chroma_list_set_font_information (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), bmp_fonts[BMP_FONT_NORMAL]);

    if (self->window)
        gtk_widget_queue_draw (self->window);

}

#if 0
static void
configuration_callback_titlestring_format (GObject *object, gpointer data)
{
	update_view (data, NULL);
}
#endif

static void
window_playlist_create_actions(BmpWindowPlaylist *self)
{
  GError *error = NULL;
  gchar *filename = NULL;

  gtk_action_group_add_actions (bmp_actions, bmp_actions_PL, G_N_ELEMENTS (bmp_actions_PL), NULL);
  gtk_action_group_set_sensitive (bmp_actions, TRUE);

  filename = g_build_filename (DATA_DIR, BMP_UI_BASENAME,"playlist.ui",NULL);

  if (!gtk_ui_manager_add_ui_from_file (bmp_ui_manager, filename, &error)) {
		g_message ("building menus failed: %s", error->message);
		g_error_free (error);
    }

#ifndef HAVE_XINE
    gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_ADD_CD), FALSE);
#endif

  g_free (filename);
}

static void
window_playlist_allocate_widgets(BmpWindowPlaylist *self)
{
#if 0
  BmpWindow *window = BMP_WINDOW(self->window);
  BmpButton *button;
  GtkAction *action;
  gint i;
#endif

  self->private->array_buttons = g_ptr_array_new();

#if 0
  /* Normal buttons */
  for (i = 0; i < nbuttons_window_playlist; ++i) {

      button = bmp_button_new(FALSE);

      if (bmp_buttons_window_playlist[i].action_id != NULL) {
             g_message("REGISTER action %s",bmp_buttons_window_playlist[i].action_id);
             action = gtk_action_group_get_action(bmp_actions, bmp_buttons_window_playlist[i].action_id);
             if (action != NULL) {
                     gtk_action_connect_proxy (GTK_ACTION (action), GTK_WIDGET (button));
             }
        }

      /* FIXME: Make this an appropriate method of BmpWindow */
      gtk_fixed_put(GTK_FIXED (BMP_WINDOW(window)->canvas), GTK_WIDGET (button), bmp_buttons_window_playlist[i].pos_x, bmp_buttons_window_playlist[i].pos_y);
      g_ptr_array_add(self->private->array_buttons,button);
  }
#endif
}

static void
adapt_cursor (BmpWindowPlaylist *self, gint x, gint y)
{
   BmpWindow *window = BMP_WINDOW(self->window);
   GdkCursor *cursor;
   GdkCursorType cursor_type = GDK_LEFT_PTR;
   gint w_width, w_height;

   w_width = GTK_WIDGET(window)->allocation.width;
   w_height = GTK_WIDGET(window)->allocation.height;

   if ( (y>=(PLEDIT_ROOT_Y+PLEDIT_INNERPAD+1)) &&
	(y<=(w_height - PLEDIT_INNER_YMARGIN)) &&
        (x>=(w_width - self->private->statvals_int[BMP_STAT_VSEP_WIDTH]) - 2) &&
        (x<=(w_width - self->private->statvals_int[BMP_STAT_VSEP_WIDTH]) + 2) &&
	(chroma_active_widget == BMP_CHROMA_WIDGET_NONE)
      )
	{
		self->private->statvals_bool[BMP_STAT_VSEP_INSIDE] = TRUE;

		cursor_type = GDK_SB_H_DOUBLE_ARROW;
		goto adapt_cursor__set_cursor;
	}
    else if ( (y>=(PLEDIT_ROOT_Y+PLEDIT_INNERPAD+1)) &&
	(y<=(w_height - PLEDIT_INNER_YMARGIN)) &&
        (x!=(w_width - self->private->statvals_int[BMP_STAT_VSEP_WIDTH])) &&
	(self->private->statvals_bool[BMP_STAT_VSEP_INSIDE] == TRUE)
	)
	{
		self->private->statvals_bool[BMP_STAT_VSEP_INSIDE] = FALSE;
		/* FIXME: Generalize this also for the other windows */
		goto adapt_cursor__set_cursor;
 	}

#if 0
   if (((w_width-x) <= 24) && ((w_height-y) <=24))
   {
   		cursor_type = GDK_BOTTOM_RIGHT_CORNER;
		goto adapt_cursor__set_cursor;
   }


   /* FIXME: Inconsistency. This should use a stat val too. */
   if ((y < 16) && (w_width-x > 20))
   	{
		cursor_type = GDK_FLEUR;
		goto adapt_cursor__set_cursor;
	}
#endif
  return;

  adapt_cursor__set_cursor:

  if (cursor_type != GDK_LEFT_PTR)
  	cursor = gdk_cursor_new_for_display (gdk_display_get_default(), cursor_type);
  else
	cursor = bmp_cursor_get (CURSOR_PLEDIT);

  gdk_window_set_cursor (GDK_WINDOW (GTK_WIDGET (BMP_WINDOW(window)->canvas)->window), cursor);
  return;
}

static gboolean
on_playlist_real_window_button_release (GtkWidget *widget,
		    GdkEventButton *event,
		    gpointer data)
{
    if (dock_is_moving(GTK_WINDOW(widget)))
   {
	dock_move_release(GTK_WINDOW(widget));
	bmp_system_control_move_end (bmp_system_control);
   }

    return FALSE;
}

static gboolean
on_window_playlist_button_release (GtkWidget *widget,
		    GdkEventButton *event,
		    BmpWindowPlaylist *self)
{
    gboolean aux_boolean = FALSE;
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    /* FIXME: The VSEP needs to be a chroma widget as well */
    if (self->private->statvals_bool[BMP_STAT_VSEP_DRAG] == TRUE)
	self->private->statvals_bool[BMP_STAT_VSEP_DRAG] = FALSE;

    adapt_cursor (self, (gint)event->x, (gint)event->y);

    if (chroma_active_widget != BMP_CHROMA_WIDGET_NONE)
                {
			switch (chroma_widgets[chroma_active_widget]->type)
	        	{
			    case BMP_CHROMA_WIDGET_TYPE_VSCROLLBAR:
			    {
				aux_boolean = chroma_vscrollbar_dispatch_button_release (CHROMA_VSCROLLBAR(chroma_widgets[chroma_active_widget]->widget), event);
	       		    }
			    break;

			    case BMP_CHROMA_WIDGET_TYPE_BUTTON:
			    {
				aux_boolean = chroma_button_dispatch_button_release (CHROMA_BUTTON(chroma_widgets[chroma_active_widget]->widget), event->x, event->y);
	       		    }
			    break;

			    case BMP_CHROMA_WIDGET_TYPE_LIST:
			    {
				aux_boolean = chroma_list_dispatch_button_release (CHROMA_LIST(chroma_widgets[chroma_active_widget]->widget), event);
	        	    }
			    break;

			    default:
				aux_boolean = FALSE;
				break;
	        	}

			chroma_active_widget = BMP_CHROMA_WIDGET_NONE;
                }

    return FALSE;
}

static gboolean
on_playlist_real_window_button_press (GtkWidget *widget,
		     GdkEventButton *event,
		     gpointer data)
{
    gint w_width, w_height;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    gtk_window_get_size (GTK_WINDOW (widget), &w_width, &w_height);

    if ((event->type == GDK_BUTTON_PRESS) && (event->button == 1) && ( (event->y < 16) && (w_width-event->x > 20)))
    {
 	dock_move_press (GTK_WINDOW(widget), event, FALSE);
	bmp_system_control_move_start (bmp_system_control);
    }

    return FALSE;
}

static gboolean
on_playlist_real_window_key_press (GtkWidget *widget,
		    GdkEventKey *event,
		    BmpWindowPlaylist *self)
{
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if (event->keyval == GDK_Escape)
    {
	bmp_window_main_iconify (RM_GET_ITEM("windows", "window-main", NOLOCK));
	return FALSE;
    }

    chroma_list_dispatch_key_press (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget), event);

    return FALSE;
}

static gboolean
on_window_playlist_button_press (GtkWidget *widget,
		    GdkEventButton *event,
		    BmpWindowPlaylist *self)
{
    GtkWidget *menu;
    GtkRequisition req;
    gint x, y, w_width, w_height, w_xpos, w_ypos, n;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    gtk_window_get_size (GTK_WINDOW (self->window), &w_width, &w_height);
    gtk_window_get_position (GTK_WINDOW (self->window), &w_xpos, &w_ypos);

    x = event->x;
    y = event->y;

    if (event->button == 3)
	{
		if ( (chroma_list_inside(CHROMA_LIST (chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), event->x, event->y)) &&
		     (! self->private->statvals_bool[BMP_STAT_VSEP_SNAPPED]))
		{
			GList *paths;

			paths =
			    chroma_list_get_selected_paths (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));
			playlists_adapt_actions (self, (paths!=NULL)?TRUE:FALSE);
			if (paths) g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);

			menu =
			    ui_manager_get_popup (bmp_ui_manager, "/popup/menu-playlists");
			gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);

			return FALSE;
		}

		if (chroma_list_inside(CHROMA_LIST (chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget), event->x, event->y))
		{
			GList *paths;

			paths =
			    chroma_list_get_selected_paths (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));

			tracklist_adapt_actions (self, CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
			tracklist_adapt_actions_sel (self, chroma_list_get_n_items(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget)) != 0);
			gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_TRACKLIST_FILEINFO), ((paths != NULL) && (g_list_length(paths)==1)));
			if (paths) g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);

			menu =
			    ui_manager_get_popup (bmp_ui_manager, "/popup/menu-tracklist");
			gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
			return FALSE;
		}

		return FALSE;
	}

    if (event->button == 1)
    {
	if (REGION_L(12, 37, 29, 11, w_width, w_height))
	{
	        /* ADD button menu */
		menu = ui_manager_get_popup (bmp_ui_manager, "/popup/menu-add");
		gtk_widget_size_request (menu, &req);
		bmp_menu_popup (GTK_MENU(menu), (w_xpos+14), (w_ypos+w_height) - 8 - req.height, event->button, event->time);
		return FALSE;
	}
	else
	if REGION_L(41, 66, 29, 11, w_width, w_height)
	{
	        /* SUB button menu */
		tracklist_adapt_actions (self, CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
		menu = ui_manager_get_popup (bmp_ui_manager, "/popup/menu-del");
		gtk_widget_size_request (menu, &req);
		bmp_menu_popup (GTK_MENU(menu), (w_xpos+43), (w_ypos+w_height) - 8 - req.height, event->button, event->time);
		return FALSE;
	}
	else
	if REGION_L(70, 95, 29, 11, w_width, w_height)
	{
	        /* SEL button menu */
		tracklist_adapt_actions_sel (self, chroma_list_get_n_items(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget)) != 0);
		menu = ui_manager_get_popup (bmp_ui_manager, "/popup/menu-select");
		gtk_widget_size_request (menu, &req);
		bmp_menu_popup (GTK_MENU(menu), (w_xpos+72), (w_ypos+w_height) - 8 - req.height, event->button, event->time);
		return FALSE;
    	}
    	else
	if REGION_L(99, 124, 29, 11, w_width, w_height)
	{
	        /* MISC button menu */
		tracklist_adapt_actions_sort (self, chroma_list_get_n_items(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget)) != 0);
		menu = ui_manager_get_popup (bmp_ui_manager, "/popup/menu-sort");
		gtk_widget_size_request (menu, &req);
		bmp_menu_popup (GTK_MENU(menu), (w_xpos+101), (w_ypos+w_height) - 8 - req.height, event->button, event->time);
		return FALSE;
	}
	else if REGION_R(46,23,29,11, w_width, w_height)
	{
		/* LIST menu */
		menu = ui_manager_get_popup (bmp_ui_manager, "/popup/menu-list");
		gtk_widget_size_request (menu, &req);
		bmp_menu_popup (GTK_MENU(menu), w_xpos + w_width - 12 - req.width, (w_ypos+w_height) - 8 - req.height, event->button, event->time);
		return FALSE;
	}
        else if (event->button == 1 && ( ((w_width-event->x) <= 24) && ((w_height-event->y) <=24)))
	{
		/*FIXME: This looks sluggish */
		gtk_window_begin_resize_drag (GTK_WINDOW(self->window),
				      GDK_WINDOW_EDGE_SOUTH_EAST,
				      event->button,
				      event->x_root,
				      event->y_root,
			              event->time);

		adapt_cursor (self, event->x, event->y);
	        return TRUE;
	}
    else  if ( (y>=(PLEDIT_ROOT_Y+PLEDIT_INNERPAD+1)) &&
               (y<=(w_height - PLEDIT_INNER_YMARGIN)) &&
               (x>=(w_width - self->private->statvals_int[BMP_STAT_VSEP_WIDTH]) - 2) &&
               (x<=(w_width - self->private->statvals_int[BMP_STAT_VSEP_WIDTH]) + 2) &&
	       (self->private->statvals_bool[BMP_STAT_VSEP_INSIDE] == TRUE)
          )
         {
	        self->private->statvals_bool[BMP_STAT_VSEP_DRAG] = TRUE;
		return FALSE;
         }
   }

   for (n = 0; n < BMP_N_CHROMA_WIDGETS; n++)
   {
        gboolean aux_boolean = FALSE;

	switch (chroma_widgets[n]->type)
        {
	    case BMP_CHROMA_WIDGET_TYPE_VSCROLLBAR:
	    {
		aux_boolean = chroma_vscrollbar_dispatch_button_press (CHROMA_VSCROLLBAR(chroma_widgets[n]->widget), event);
            }
	    break;

	    case BMP_CHROMA_WIDGET_TYPE_LIST:
	    {
		aux_boolean = chroma_list_dispatch_button_press (CHROMA_LIST(chroma_widgets[n]->widget), event);
            }
	    break;

	    default:
		aux_boolean = FALSE;
		break;

        }

   	if (aux_boolean)
        {
	    gtk_widget_queue_draw (self->window);
	    chroma_active_widget = n;
	    break;
        }
   }

    return FALSE;
}

static gboolean
on_playlist_real_window_motion_notify (GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
    if (dock_is_moving(GTK_WINDOW(widget)))
	dock_move_motion(GTK_WINDOW(widget), event);

    return FALSE;
}

static gboolean
on_window_playlist_motion_notify (GtkWidget *widget, GdkEventMotion *event, BmpWindowPlaylist *self)
{
    gint x, y, w_width, w_height;
    GdkModifierType state;

#if 0
    if (event->is_hint)
      {
#endif
	x = event->x;
        y = event->y;
        state = event->state;

        gdk_window_get_pointer (event->window, &x, &y, &state);

        w_width = GTK_WIDGET(self->window)->allocation.width;
        w_height = GTK_WIDGET(self->window)->allocation.height;

	if (!self->private->statvals_bool[BMP_STAT_VSEP_DRAG])
	{
	        gboolean aux_boolean = FALSE;
		adapt_cursor (self, x, y);

		if (chroma_active_widget != BMP_CHROMA_WIDGET_NONE)
                {
			switch (chroma_widgets[chroma_active_widget]->type)
	        	{
			    case BMP_CHROMA_WIDGET_TYPE_VSCROLLBAR:
			    {
				aux_boolean = chroma_vscrollbar_dispatch_motion (CHROMA_VSCROLLBAR(chroma_widgets[chroma_active_widget]->widget),event);
	       		    }
			    break;

			    case BMP_CHROMA_WIDGET_TYPE_LIST:
			    {
				aux_boolean = chroma_list_dispatch_motion (CHROMA_LIST(chroma_widgets[chroma_active_widget]->widget), event);
	        	    }
			    break;

			    default:
				aux_boolean = FALSE;
				break;
	        	}
                }
	}
	else
	{
		/* Sanity checks for VSEP position */

		if ( ((w_width - event->x) < VSEP_MIN_WIDTH) && ((w_width - event->x) > VSEP_SNAP_WIDTH))
		    return FALSE;

		if  ( ((w_width - event->x) >= VSEP_MIN_WIDTH) && (!self->private->statvals_bool[BMP_STAT_VSEP_SNAPPED]))
		{
			self->private->statvals_int[BMP_STAT_VSEP_WIDTH] = w_width - event->x;
			bmp_window_playlist_resize_widgets (self);
			gtk_widget_queue_draw (self->window);
		}
		else if ((w_width - event->x >= VSEP_SNAP_WIDTH))
		{
			self->private->statvals_int[BMP_STAT_VSEP_WIDTH] =  VSEP_MIN_WIDTH;
			self->private->statvals_bool[BMP_STAT_VSEP_SNAPPED] = FALSE;
			bmp_window_playlist_resize_widgets (self);
			gtk_widget_queue_draw (self->window);
		}
		else if ((w_width - event->x) < VSEP_MIN_WIDTH)
		{
			self->private->statvals_bool[BMP_STAT_VSEP_SNAPPED] = TRUE;
			self->private->statvals_int[BMP_STAT_VSEP_WIDTH] = VSEP_SNAP_WIDTH;
			bmp_window_playlist_resize_widgets (self);
			gtk_widget_queue_draw (self->window);
		}

		if ((w_width - event->x) > (w_width/2))
		{
			self->private->statvals_int[BMP_STAT_VSEP_WIDTH] = w_width / 2;
			bmp_window_playlist_resize_widgets (self);
			gtk_widget_queue_draw (self->window);
		}

        }
#if 0
      }
    else
      {
        x = event->x;
        y = event->y;
        state = event->state;
     }
#endif

  return FALSE;
}

/* DnD */
static void
on_window_playlist_drag_end (GtkWidget *widget,
			     GdkDragContext *context,
			     gpointer user_data)
{
	chroma_list_dispatch_drag_leave(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
	chroma_list_dispatch_drag_leave(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));

	gtk_widget_queue_draw (GTK_WIDGET(user_data));
}

static void
on_window_playlist_drag_leave (GtkWidget *widget,
				GdkDragContext *context,
				guint time,
				gpointer user_data)
{
	chroma_list_dispatch_drag_leave(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
	chroma_list_dispatch_drag_leave(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));

	gtk_widget_queue_draw (GTK_WIDGET(user_data));
}


static void
on_window_playlist_drag_data_received(GtkWidget * widget,
                                      GdkDragContext *context,
                                      gint x, gint y,
                                      GtkSelectionData *
                                      selection_data, guint info,
                                      guint time, BmpWindowPlaylist *self)
{
    ChromaList *list;

    g_return_if_fail(selection_data != NULL);

    /* accept the drop */
    if (!selection_data->data)
      {
          g_message("Received no DND data!");
  	  gtk_drag_finish (context, FALSE, FALSE, time);
          return;
      }

    list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
    if ((chroma_list_inside(list, x, y)) && chroma_list_get_list (list))
    {
	  gchar *uri_string;
  	  gchar **uris;
	  gint ctr = chroma_list_get_drop_position(list);
	  gint new_ctr = 0;

	  uri_string = g_strdup((gchar*)selection_data->data);
	  g_strstrip(uri_string);

	  uris = g_uri_list_extract_uris (uri_string);

	  bmp_system_control_add_uri_list (bmp_system_control,
			uris,
			-1,
			ctr,
			&new_ctr,
			NULL);

	  gtk_drag_finish (context, TRUE, FALSE, time);
	  chroma_list_dispatch_drag_leave(list);
#if 0
	FIXME: ?
	  if (uris) g_strfreev(uris);
#endif
	  return;
    }

    list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget);
    if (chroma_list_inside(list, x, y)) {

    	gchar *uri_string, *basename = NULL;
	gchar **uris;
	gint ctr = 0, new_ctr = 0;
	gint tracklist;
	GURI *g_uri;

	uri_string =
	    g_strdup((gchar*)selection_data->data);
	g_strstrip(uri_string);

	uris =
	    g_uri_list_extract_uris (uri_string);

	while (*uris)
	  {

	    g_uri = bmp_uri_new (*uris);
	    bmp_uri_unescape (g_uri);

	    if (g_file_test (g_uri->path, G_FILE_TEST_IS_DIR))
              {
		basename =
		    g_path_get_basename(g_uri->path);
		tracklist =
		    bmp_playlist_tracklist_new (bmp_playlist, basename);

	        if (tracklist != -1)
                  {
		    gchar *uris_tmp[2];
	
		    uris_tmp[0] =
			*uris;	
		    uris_tmp[1] =
			NULL;

		    bmp_system_control_add_uri_list (bmp_system_control, uris_tmp, tracklist, -1, &new_ctr, NULL);
		  }
                else
                  {
		    g_message ("%s: This application thinks that '%s' is not a directory", G_STRLOC, g_uri->path);
		  }
	      }

	    bmp_uri_delete (g_uri);

	    *uris++;
	    ctr++;
	}

	if (ctr == 1)
	    chroma_list_select_name (list, basename);

	gtk_drag_finish (context, TRUE, FALSE, time);
	chroma_list_dispatch_drag_leave(list);
#if 0
    FIXME: ?
        if (uris) g_strfreev(uris);
#endif
	return;
    }
}

static gboolean
on_window_playlist_drag_motion (GtkWidget *widget,
				GdkDragContext *context,
				gint x,
				gint y,
				guint time,
				BmpWindowPlaylist *self)

{
	ChromaList *list_playlists, *list_tracklist;
	GdkAtom target;

	target = gtk_drag_dest_find_target (widget, context, NULL);
	if (target == GDK_NONE)
	{
	    gdk_drag_status (context, 0, time);
	    return FALSE;
	}

	list_tracklist = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
	list_playlists = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget);

	if ((chroma_list_get_list(list_tracklist)) &&
	     (chroma_list_inside(list_tracklist, x, y)))
	{
		chroma_list_dispatch_drag_leave(list_playlists);
		chroma_list_dispatch_drag_motion(list_tracklist, x, y);
		gtk_widget_queue_draw (self->window);
		gdk_flush();
		return FALSE;
	}
	else
	if  (chroma_list_inside(list_playlists, x, y))
	{
		chroma_list_dispatch_drag_leave(list_tracklist);
		chroma_list_dispatch_drag_motion(list_playlists, x, y);
		gtk_widget_queue_draw (self->window);
		gdk_flush();
		return FALSE;
	}
	else
	{
		chroma_list_dispatch_drag_leave(list_tracklist);
		chroma_list_dispatch_drag_leave(list_playlists);
		gdk_drag_status (context, 0, time);
		gtk_widget_queue_draw (self->window);
		gdk_flush();
		return FALSE;
	}

	return FALSE;
}
/* END Dnd */

void
on_chroma_widget_expose_request (GObject *object, gint pos_x, gint pos_y, gint width, gint height, gpointer data)
{
	/*FIXME: This is sort of pointless since the expose handler redraws the whole list everytime.
		 We should break this down into
 		 widget components instead of into coordinates,
		 and make the expose handler draw only the relevant widget */

	    gtk_widget_queue_draw_area (GTK_WIDGET(data),
						pos_x,
					  	pos_y,
						width,
						height);
}

static void
resize_widgets (BmpWindowPlaylist *self, gint w, gint h)
{
        ChromaList *list;
	ChromaVScrollbar *vs;
	gint vsep_width = 0, offset = 0, h_offset = 0, n_items, n_item_display; 

	vsep_width =
	    self->private->statvals_int[BMP_STAT_VSEP_WIDTH];
	offset =
	    PLEDIT_ROOT_Y + bmp_fonts[BMP_FONT_LARGE]->font_height + 14;

	/* List Playlists */
	list =
	    CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget);
        n_items =
	    g_object_get_int (list, "n_items");
        n_item_display =
	    chroma_list_get_n_item_display (list);
	h_offset =
	    ((!(n_items-n_item_display) ? 0 : (n_items/n_item_display))) ? VSCROLLBAR_DEFAULT_WIDTH : 0;

	chroma_list_move_resize (list,
			w - vsep_width + 8,
		        PLEDIT_ROOT_Y + PLEDIT_INNERPAD + 4, 
			vsep_width - PLEDIT_INNER_XMARGIN - h_offset - 8,
		        h - PLEDIT_ROOT_Y - PLEDIT_INNER_YMARGIN);

	/* Scrollbar Playlists */
	vs = CHROMA_VSCROLLBAR(chroma_widgets[BMP_CHROMA_VSCROLLBAR_PLAYLISTS]->widget);
	chroma_vscrollbar_move_resize (vs,
				       w - PLEDIT_INNER_XMARGIN - VSCROLLBAR_DEFAULT_WIDTH - 4,
				       PLEDIT_ROOT_Y + PLEDIT_INNERPAD + 4,
				       VSCROLLBAR_DEFAULT_WIDTH,
				       h - PLEDIT_INNER_YMARGIN - PLEDIT_ROOT_Y);

	/* List Tracklist */
	list =
	    CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
        n_items =
	    g_object_get_int (list, "n_items");
        n_item_display =
	    chroma_list_get_n_item_display (list);
	h_offset =
	    ((!(n_items-n_item_display) ? 0 : (n_items/n_item_display))) ? VSCROLLBAR_DEFAULT_WIDTH + 2 : 0;

	chroma_list_move_resize (list,
				 PLEDIT_ROOT_X + 4,
				 offset,
				 w - (vsep_width + PLEDIT_ROOT_X + 10) - h_offset,
				 (h - PLEDIT_INNER_YMARGIN) - offset + 8);

       /* Scrollbar Tracklist */
       vs = CHROMA_VSCROLLBAR(chroma_widgets[BMP_CHROMA_VSCROLLBAR_TRACKLIST]->widget);
       chroma_vscrollbar_move_resize (vs,
				      w - (vsep_width + PLEDIT_ROOT_X) - (VSCROLLBAR_DEFAULT_WIDTH - 7),
				      offset,
				      VSCROLLBAR_DEFAULT_WIDTH,
				      (h - PLEDIT_INNER_YMARGIN) - offset + 8);

       gtk_widget_queue_draw (self->window);
}

static gboolean
on_window_playlist_configure (GtkWidget *widget, GdkEventConfigure *event, BmpWindowPlaylist *self)
{
        static gint prev_width = -1, prev_height = -1;

        g_return_val_if_fail (GTK_IS_WIDGET(widget), FALSE);

	if (prev_width != -1)
	{
		if ((prev_width != event->width) || (prev_height != event->height))
	        {
		    resize_widgets (self, event->width, event->height);
		    gtk_widget_queue_draw (self->window);
		}
	}

	prev_width = event->width;
	prev_height = event->height;

	return FALSE;
}

static void
on_window_playlist_scroll(GtkWidget *widget, GdkEventScroll *event, gpointer data)
{
	if(event->type == GDK_SCROLL){
		GObject *object;

		guint x,y,n;
		x = event->x;
		y = event->y;


		for(n=0;n<BMP_N_CHROMA_WIDGETS;n++){
			gboolean aux_boolean = FALSE;

			switch(chroma_widgets[n]->type){
				case BMP_CHROMA_WIDGET_TYPE_VSCROLLBAR:
					aux_boolean =
						chroma_vscrollbar_dispatch_scroll(CHROMA_VSCROLLBAR(chroma_widgets[n]->widget), event);
					break;
				case BMP_CHROMA_WIDGET_TYPE_LIST:
					if (!chroma_list_inside(CHROMA_LIST (chroma_widgets[n]->widget), x, y)) break;
					object = g_object_get_data (chroma_widgets[n]->widget, "adjustment");
					aux_boolean =
						chroma_vscrollbar_dispatch_scroll_direct(CHROMA_VSCROLLBAR(object), event);
					break;

				default:
					aux_boolean = FALSE;
					break;
			}
		}
	}
}

static void
tracklist_adapt_actions (BmpWindowPlaylist *self, ChromaList *list)
{
	GList *paths;

	paths = chroma_list_get_selected_paths (list);

	if (!paths) {
	    tracklist_adapt_actions_has_sel (self, FALSE);
	} else {
	    tracklist_adapt_actions_has_sel (self, TRUE);
	    g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	}

	tracklist_adapt_actions_has_items (self, (chroma_list_get_n_items (list) == 0) ? FALSE : TRUE);
}

static void
on_playlist_tracklist_changed (ChromaList *list, BmpWindowPlaylist *self)
{
	tracklist_adapt_actions (self, list);

	if (self->window != NULL)
		bmp_window_playlist_resize_widgets (self);
}

static void
on_system_control_current_tracklist_set (BmpSystemControl *playlist, gint _idx, BmpWindowPlaylist *self)
{
    ChromaList *list_tracklist, *list_playlists;
    BmpListType *tracklist;

    if (_idx < 0) {
	playlists_adapt_actions_changed (self, FALSE);
	return;
    }

    list_tracklist = CHROMA_LIST (chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
    list_playlists = CHROMA_LIST (chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget);

    tracklist = bmp_playlist_tracklist_get_i (bmp_playlist, _idx); 

    if ((chroma_list_get_list (list_tracklist) != tracklist) || (tracklist == NULL))
      {
	    chroma_list_set_list (list_tracklist, tracklist);
	    playlists_adapt_actions_changed (self, TRUE);

	    /* FIXME: Cross-playlist history doesn't work yet */
	    bmp_system_control_clear_playback_history (bmp_system_control, NULL);
	    gtk_action_set_sensitive (gtk_action_group_get_action(bmp_actions, BMP_ACTION_CLEAR_PLAYBACK_HISTORY), FALSE);
      }

    if (self->window != NULL)
	bmp_window_playlist_resize_widgets (self);
}

static void
on_playlist_playlists_changed (ChromaList *list, BmpWindowPlaylist *self) {

    GList *paths;
    gint n;

    paths =
	chroma_list_get_selected_paths (list);

    if (!paths)
      {
	bmp_system_control_current_tracklist_set (bmp_system_control, -1);
	playlists_adapt_actions_changed (self, FALSE);
      }
    else
      {
	n = gtk_tree_path_get_indices(paths->data)[0];
	bmp_system_control_current_tracklist_set (bmp_system_control, n);
	g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
      }

    bmp_ui_callback_fileinfo_update();
}

static void
on_playlist_tracklist_row_activated (ChromaList *list, gint _idx, BmpPlay *play) {

	bmp_system_control_play_list_track (bmp_system_control,
					    g_object_get_int (bmp_system_control, "current_tracklist"),
					    _idx);
}

static void
on_playlist_playlists_row_activated (ChromaList *list, gint _idx, BmpPlay *play)
{
	bmp_system_control_play_list_track (bmp_system_control,
					    _idx,
					    0);
}

static void
playlist_callback_list_deleted (BmpSystemControl *system_control, gint list, BmpWindowPlaylist *self)
{
    if ((list == g_object_get_int(system_control,"current_tracklist")) || (list == -1))
      {
	chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));
	chroma_list_set_list (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget), NULL);
      }

    bmp_window_playlist_resize_widgets (self);
    gtk_widget_queue_draw (GTK_WIDGET(self->window));
}

static void
playlist_callback_list_added (BmpSystemControl *system_control, gint list, BmpWindowPlaylist *self)
{
	chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));

	bmp_window_playlist_resize_widgets (self);
	gtk_widget_queue_draw (GTK_WIDGET(self->window));
}

static void
playlist_callback_list_renamed (BmpSystemControl *system_control, gint list, BmpWindowPlaylist *self)
{
	chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));
	bmp_window_playlist_resize_widgets (self);
	gtk_widget_queue_draw (GTK_WIDGET(self->window));
}

static void
playlist_callback_tracklist_list_sorted (BmpSystemControl *system_control, gint list, BmpWindowPlaylist *self)
{
	if (list == g_object_get_int (bmp_system_control, "current_tracklist"))
          {
	    chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
	    bmp_window_playlist_resize_widgets (self);
	    gtk_widget_queue_draw (GTK_WIDGET(self->window));
          }
}

static void
playlist_callback_tracklist_items_removed (BmpSystemControl *system_control, gint list, GArray *removed_rows, gint n_removed_rows, BmpWindowPlaylist *self)
{
	if (list == g_object_get_int (bmp_system_control, "current_tracklist"))
          {
	    chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
	    bmp_window_playlist_resize_widgets (self);
	    gtk_widget_queue_draw (GTK_WIDGET(self->window));
          }
}

static void
playlist_callback_tracklist_items_added (BmpSystemControl *system_control, gint list, GArray *added_rows, gint n_added_rows, BmpWindowPlaylist *self)
{
	if (list == g_object_get_int (bmp_system_control, "current_tracklist"))
          {
	    chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
	    bmp_window_playlist_resize_widgets (self);
	    gtk_widget_queue_draw (GTK_WIDGET(self->window));
          }
}

static void
set_clear_history_sensitive (BmpSystemControl *system_control, gpointer data)
{
    gtk_action_set_sensitive (GTK_ACTION(gtk_action_group_get_action (bmp_actions, BMP_ACTION_CLEAR_PLAYBACK_HISTORY)), TRUE);
}

static void
on_chroma_list_tracklist_cursor_position (ChromaList *list, gint position, BmpWindowPlaylist *self) 
{
    tracklist_adapt_actions (self, list);
}

static GdkFilterReturn
_window_filter_func (GdkXEvent *xevent,
                                             GdkEvent *event,
                                             gpointer data)
{
    g_message ("%s: WINDOW, XEVENT TYPE: %d", G_STRLOC, ((XEvent*)xevent)->type);

    return GDK_FILTER_CONTINUE;
}

static GdkFilterReturn
_canvas_filter_func (GdkXEvent *xevent,
                                             GdkEvent *event,
                                             gpointer data)
{
    g_message ("%s: CANVAS, XEVENT TYPE: %d", G_STRLOC, ((XEvent*)xevent)->type);

    return GDK_FILTER_CONTINUE;
}


static void
window_playlist_create(BmpWindowPlaylist *self) {

    BmpWindow *window;
    ChromaList *list_tracklist, *list_playlists;
    ChromaVScrollbar *vs_tracklist, *vs_playlists;
    gint w, h, n;

    self->window =
	GTK_WIDGET(bmp_window_new(GTK_WINDOW_TOPLEVEL));
    window_playlist_set_geometry_hints(self);

    window =
	BMP_WINDOW(self->window);

    gtk_widget_realize (self->window);
    gtk_widget_realize (BMP_WINDOW(self->window)->canvas);

#if 0
    gdk_window_add_filter (GTK_WIDGET(self->window)->window, _window_filter_func, NULL);
    gdk_window_add_filter (GTK_WIDGET(BMP_WINDOW(self->window)->canvas)->window, _canvas_filter_func, NULL);
#endif

    window_playlist_create_actions (self);
    window_playlist_allocate_widgets (self);

    bmp_window_set_icon_list (GTK_WIDGET(window), "playlist");
    gtk_window_set_title (GTK_WINDOW(window), "BMP Playlist");

    g_object_set_data (G_OBJECT(window), "id", g_strdup("playlist"));

    dock_window_add (GTK_WINDOW(window));

    for (n = 0; n < BMP_N_CHROMA_WIDGETS; chroma_widgets[n++] = g_new0(BmpChromaWidget,1));

    vs_playlists   = CHROMA_VSCROLLBAR(chroma_vscrollbar_new ());
    chroma_widgets[BMP_CHROMA_VSCROLLBAR_PLAYLISTS]->widget   = G_OBJECT(vs_playlists);
    chroma_widgets[BMP_CHROMA_VSCROLLBAR_PLAYLISTS]->type     = BMP_CHROMA_WIDGET_TYPE_VSCROLLBAR;

    vs_tracklist   = CHROMA_VSCROLLBAR(chroma_vscrollbar_new ());
    chroma_widgets[BMP_CHROMA_VSCROLLBAR_TRACKLIST]->widget   = G_OBJECT(vs_tracklist);
    chroma_widgets[BMP_CHROMA_VSCROLLBAR_TRACKLIST]->type     = BMP_CHROMA_WIDGET_TYPE_VSCROLLBAR;


    list_playlists = CHROMA_LIST(chroma_list_new (GTK_WIDGET(window), 0, 0, 0, 0));
    chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget         = G_OBJECT(list_playlists);
    chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->type           = BMP_CHROMA_WIDGET_TYPE_LIST;
    chroma_list_set_align(list_playlists, CHROMA_LIST_ALIGN_LEFT);
    chroma_list_set_selection_mode(list_playlists, GTK_SELECTION_SINGLE);
    chroma_list_set_drop_mode(list_playlists, CHROMA_LIST_DROP_WHOLE);
    chroma_list_set_change_on_select(list_playlists, TRUE);

    list_tracklist =  CHROMA_LIST(chroma_list_new (GTK_WIDGET(window), 0, 0, 0, 0));
    chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget         = G_OBJECT(list_tracklist);
    chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->type           = BMP_CHROMA_WIDGET_TYPE_LIST;
    chroma_list_set_align(list_tracklist, CHROMA_LIST_ALIGN_LEFT);
    chroma_list_set_columns(list_tracklist, 2);
    g_object_connect (G_OBJECT(list_tracklist), 
		      "signal::cursor-position", 
		      G_CALLBACK (bmp_ui_callback_fileinfo_update), 
		      NULL, 
		      "signal::cursor-position", 
		      G_CALLBACK (on_chroma_list_tracklist_cursor_position), 
		      self, 
		      NULL);

    /* Tracklist */
    chroma_list_set_adjustment (list_tracklist, vs_tracklist);
    g_object_set_data (G_OBJECT(vs_tracklist), "list", list_tracklist);

    /* Playlist */
    chroma_list_set_adjustment (list_playlists, vs_playlists);
    g_object_set_data (G_OBJECT(vs_playlists), "list", list_playlists);

    for (n = 0; n < BMP_N_CHROMA_WIDGETS;
       g_signal_connect (G_OBJECT(chroma_widgets[n++]->widget), "expose_request", G_CALLBACK(on_chroma_widget_expose_request), BMP_WINDOW (window)->canvas));

    /* Inactivate all Chroma widgets */
    chroma_active_widget = BMP_CHROMA_WIDGET_NONE;

    self->private->statvals_int[BMP_STAT_VSEP_WIDTH] =  VSEP_MIN_WIDTH;
    self->private->statvals_bool[BMP_STAT_VSEP_INSIDE] = FALSE;

    /* DnD */
    bmp_drag_dest_set (GTK_WIDGET(window->canvas));
    g_object_connect (G_OBJECT(BMP_WINDOW(window)->canvas),
			"signal::drag-leave",
			G_CALLBACK(on_window_playlist_drag_leave),
			window,
			"signal::drag-data-received",
			G_CALLBACK(on_window_playlist_drag_data_received),
			self,
			"signal::drag-motion",
			G_CALLBACK(on_window_playlist_drag_motion),
			self,
			"signal::drag-end",
			G_CALLBACK(on_window_playlist_drag_end),
			window,
			NULL);

    /* BmpWindow Events */
    g_object_connect (G_OBJECT(BMP_WINDOW(window)),
			"signal::realize",
			G_CALLBACK(bmp_window_playlist_resize_widgets),
			self,
			"signal::key_press_event",
			G_CALLBACK(on_playlist_real_window_key_press),
			window,
			"signal::button_press_event",
			G_CALLBACK(on_playlist_real_window_button_press),
			window,
			"signal::button_release_event",
			G_CALLBACK(on_playlist_real_window_button_release),
			window,
			"signal::motion_notify_event",
			G_CALLBACK(on_playlist_real_window_motion_notify),
			window,
			"signal::scroll_event",
			G_CALLBACK(on_window_playlist_scroll),
			NULL,
			"signal::configure_event",
			G_CALLBACK(on_window_playlist_configure),
			self,
			"signal::focus-in-event",
			G_CALLBACK(on_window_playlist_focus_in),
			self,
			"signal::focus-out-event",
			G_CALLBACK(on_window_playlist_focus_out),
			self,
			"signal::window-state-event",
			G_CALLBACK(on_window_playlist_state_event),
			self,
			"signal::delete-event",
			G_CALLBACK(on_window_playlist_delete_event),
			self,
			NULL);

    /* BmpWindow->canvas Events */
    g_object_connect (G_OBJECT(BMP_WINDOW(window)->canvas),
			"signal::visibility-notify-event",
			G_CALLBACK(on_window_playlist_visibility),
			self,
			"signal::expose_event",
			G_CALLBACK(on_window_playlist_canvas_expose),
			self,
			"signal::button_press_event",
			G_CALLBACK(on_window_playlist_button_press),
			self,
			"signal::button_release_event",
			G_CALLBACK(on_window_playlist_button_release),
			self,
			"signal::motion_notify_event",
			G_CALLBACK(on_window_playlist_motion_notify),
			self,
			NULL);

    g_object_connect (G_OBJECT(bmp_system_control),
	    	      "signal::track_change",
		      G_CALLBACK(set_clear_history_sensitive),
		      self,
	    	      "swapped-signal::track_change",
 		      G_CALLBACK(chroma_list_mark_dirty),
                      chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget,
		      "swapped-signal::track_change",
 		      G_CALLBACK(chroma_list_mark_dirty),
                      chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget,
		      "swapped-signal::track_change",
 		      G_CALLBACK(gtk_widget_queue_draw),
                      self->window,
		      "signal::set-current-tracklist",
		      G_CALLBACK(on_system_control_current_tracklist_set),
		      self,
		      NULL);

    /* Various watches */
    g_object_connect(G_OBJECT(bmp_configuration),
		"signal::changed::bmp-font",
		G_CALLBACK(on_bmp_font_changed),
		self,
		"swapped-signal::changed::bmp-font",
		G_CALLBACK(bmp_window_playlist_resize_widgets),
		self,
		"swapped-signal::changed::bmp-font",
		G_CALLBACK(update_view),
		self,
		"swapped-signal::changed::bmp-titlestring-format",
		G_CALLBACK(update_view),
		self,
		"swapped-signal::changed::bmp-skin",
		G_CALLBACK(update_view),
		self,
		NULL);

    g_object_connect(G_OBJECT(bmp_system_control),
		"signal::list-deleted",
		G_CALLBACK(playlist_callback_list_deleted),
	 	self,
		"signal::list-added",
		G_CALLBACK(playlist_callback_list_added),
	 	self,
		"signal::list-renamed",
		G_CALLBACK(playlist_callback_list_renamed),
		self,
		"signal::tracklist-list-sorted",
		G_CALLBACK(playlist_callback_tracklist_list_sorted),
	 	self,
		"signal::tracklist-items-removed",
		G_CALLBACK(playlist_callback_tracklist_items_removed),
		self,
		"signal::tracklist-items-added",
		G_CALLBACK(playlist_callback_tracklist_items_added),
		self,
		NULL);

    gtk_window_add_accel_group (GTK_WINDOW(window), gtk_ui_manager_get_accel_group(bmp_ui_manager));

    self->private->statvals_bool[BMP_STAT_VSEP_DRAG] = FALSE;
    self->private->statvals_bool[BMP_STAT_VSEP_SNAPPED] = FALSE;

    LOOKUP_OR_CREATE_DEFAULT_INT("playlist-window","vsep-width", VSEP_MIN_WIDTH)
    set_vsep_width (self,
		bmp_configuration_key_fetch_int
			(bmp_configuration,
			"playlist-window",
			"vsep-width"));

    w = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "width");
    h = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "height");
    resize_widgets (self, w, h);

    gtk_action_set_sensitive (GTK_ACTION (gtk_action_group_get_action (bmp_actions, 
				    BMP_ACTION_CLEAR_PLAYBACK_HISTORY)), FALSE);

    gtk_widget_queue_draw (self->window);
}

/* GObject */
static void
bmp_window_playlist_init (BmpWindowPlaylist * self)
{
    self->private = g_new (BmpWindowPlaylistPrivate, 1);

    self->private->dispose_has_run = FALSE;
    self->private->ui_components = NULL;
    self->window = NULL;
    self->private->visibility = GDK_VISIBILITY_UNOBSCURED;

    window_playlist_create(self);
}

static GObject *
window_playlist_constructor (GType type,
	                     guint n_construct_properties,
         	             GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        /*
         * Invoke parent constructor.
         */
        BmpWindowPlaylistClass *klass;
        GObjectClass *parent_class;
        klass = BMP_WINDOW_PLAYLIST_CLASS (g_type_class_peek (BMP_TYPE_WINDOW_PLAYLIST));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }

    /*
     * do stuff.
     */

    return obj;
}

static void
window_playlist_dispose (GObject *gobject)
{
    BmpWindowPlaylist *self = (BmpWindowPlaylist *) gobject;

    if (self->private->dispose_has_run) {
         /*
          * If dispose did already run, return.
          */
         return;
     }

    /*
     * Make sure dispose does not run twice.
     */
    self->private->dispose_has_run = TRUE;

    /* Save playlist window states */
    bmp_configuration_key_set_boolean(bmp_configuration, "playlist-window", "visible",
	    gtk_toggle_action_get_active (GTK_TOGGLE_ACTION(gtk_action_group_get_action (bmp_toggle_actions, BMP_TOGGLE_ACTION_PL))));

    bmp_configuration_key_set_int(bmp_configuration, "playlist-window", "vsep-width", self->private->statvals_int[BMP_STAT_VSEP_WIDTH]);
}


static void
bmp_window_playlist_class_init (BmpWindowPlaylistClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->finalize = bmp_play_finalize;
#endif
    gobject_class->dispose = window_playlist_dispose;
    gobject_class->constructor = window_playlist_constructor;
}

/* Public API */
void
bmp_window_playlist_resize_widgets (BmpWindowPlaylist *self)
{
    gint w,h;

    w = self->window->allocation.width;
    h = self->window->allocation.height;

    resize_widgets (self, w, h);
}

void
bmp_window_playlist_configure(BmpWindowPlaylist *self)
{
	GtkWidget *window = self->window;
	BmpUIComponent *uic = NULL;
	gint i = 0;

	if (self->private->ui_components != NULL)
		g_hash_table_destroy (self->private->ui_components);

        self->private->ui_components = g_hash_table_new_full  (g_str_hash,
                                             g_str_equal,
                                             NULL,
                                             destroy_value);

        for (i = 0; i < G_N_ELEMENTS (bmp_playlist_component_data); i++) {

		uic = g_malloc(sizeof(BmpUIComponent));

		uic->name = 0;
  		uic->pixbuf = (GdkPixbuf*)bmp_pixbuf_new_simple (
						    bmp_playlist_component_data[i].width,
						    bmp_playlist_component_data[i].height);
		bmp_pixbuf_copy_area_simple (
					      uic->pixbuf,
					      bmp_playlist_component_data[i].x_origin,
					      bmp_playlist_component_data[i].y_origin,
					      bmp_playlist_component_data[i].width,
					      bmp_playlist_component_data[i].height);


		uic->x_origin = bmp_playlist_component_data[i].x_origin;
		uic->y_origin = bmp_playlist_component_data[i].y_origin;
		uic->width = bmp_playlist_component_data[i].width;
		uic->height = bmp_playlist_component_data[i].height;

		g_hash_table_insert	(self->private->ui_components,
					 bmp_playlist_component_data[i].name,
					 (BmpUIComponent *) uic);

	}

#if 0
	if (self->private->array_buttons != NULL) {
	        for (i = 0; i < nbuttons_window_playlist; ++i)
		{
	            BmpButton *button;

		    button = g_ptr_array_index(self->private->array_buttons, i);
		    bmp_button_set_from_table(BMP_BUTTON (button), NULL, bmp_skin_component[bmp_buttons_window_playlist[i].pixmap_id]->pixbuf, bmp_buttons_window_playlist[i].width, bmp_buttons_window_playlist[i].height, bmp_buttons_window_playlist[i].img_x, bmp_buttons_window_playlist[i].img_y, bmp_buttons_window_playlist[i].img_x_pressed, bmp_buttons_window_playlist[i].img_y_pressed);
		    gtk_widget_set_size_request (GTK_WIDGET (button), bmp_buttons_window_playlist[i].width, bmp_buttons_window_playlist[i].height);
	        }
	}
#endif

        gtk_widget_modify_bg (GTK_WIDGET(window), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));
        gtk_widget_modify_bg (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));

        gtk_widget_modify_fg (GTK_WIDGET(window), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL));
        gtk_widget_modify_fg (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL));

        gtk_widget_modify_text (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL));
	gtk_widget_modify_base (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));
        gtk_widget_modify_text (GTK_WIDGET(window), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL));
	gtk_widget_modify_base (GTK_WIDGET(window), GTK_STATE_NORMAL, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));

        gtk_widget_modify_text (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_PRELIGHT, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));
	gtk_widget_modify_base (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_PRELIGHT, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));
        gtk_widget_modify_bg (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_PRELIGHT, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));
	gtk_widget_modify_fg (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_PRELIGHT, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));

	gtk_widget_modify_base (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_SELECTED, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_SELECTED));
	gtk_widget_modify_base (GTK_WIDGET(window), GTK_STATE_SELECTED, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_SELECTED));

	gtk_widget_modify_text (GTK_WIDGET(BMP_WINDOW(window)->canvas), GTK_STATE_SELECTED, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL));
	gtk_widget_modify_text (GTK_WIDGET(window), GTK_STATE_SELECTED, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL));

	chroma_list_set_font_information (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget), bmp_fonts[BMP_FONT_NORMAL]);
	chroma_list_set_font_information (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), bmp_fonts[BMP_FONT_NORMAL]);

	chroma_list_set_use_numbers (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget),
		    FALSE);

	chroma_list_set_use_numbers (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget),
		    bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "display-tracklist-numbers"));

	SET_CURSOR_FOR_WIDGET (CURSOR_PLEDIT, BMP_WINDOW(window)->canvas);
	SET_CURSOR_FOR_WIDGET (CURSOR_PLEDIT, BMP_WINDOW(window));

	chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));
	chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));

}

gint
bmp_window_playlist_get_vsep_width (BmpWindowPlaylist *self)
{
   return self->private->statvals_int[BMP_STAT_VSEP_WIDTH];
}

static void
set_vsep_width (BmpWindowPlaylist *self, gint width)
{
   g_return_if_fail (BMP_IS_WINDOW_PLAYLIST (self));

   self->private->statvals_int[BMP_STAT_VSEP_WIDTH] = width;
   if (width <= VSEP_SNAP_WIDTH)
   {
	self->private->statvals_bool[BMP_STAT_VSEP_SNAPPED] = TRUE;
	self->private->statvals_int[BMP_STAT_VSEP_WIDTH] = VSEP_SNAP_WIDTH;
   }
   gtk_widget_queue_draw (self->window);
}

BmpWindowPlaylist*
bmp_window_playlist_new (void)
{
    BmpWindowPlaylist *self;

    self =
        g_object_new (BMP_TYPE_WINDOW_PLAYLIST, NULL);

    g_object_connect (G_OBJECT(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget),
			"signal::changed",
			G_CALLBACK(on_playlist_playlists_changed),
			self,
			"signal::row_activated",
			G_CALLBACK(on_playlist_playlists_row_activated),
			bmp_play,
			NULL);

    g_object_connect (G_OBJECT(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget),
			"signal::changed",
			G_CALLBACK(on_playlist_tracklist_changed),
			self,
			"signal::row_activated",
			G_CALLBACK(on_playlist_tracklist_row_activated),
			bmp_play,
			NULL);

    /* Initialize Actions (indirectly) */
    chroma_list_emit_changed(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));
    chroma_list_emit_changed(CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));

    return self;
}
