#include <config.h>

#include <stdio.h>
#ifndef __USE_BSD
# define __USE_BSD	/* Needed for mkdtemp () */
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <glib/gstdio.h>

#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#include "main.h"
#include "bmp_configuration.h"

#include "bmpx/util.h"
#include "bmpx/uri.h"
#include "bmpx/vfs.h"
#include "bmpx/plugin.h"
#include "bmpx/plugin_interfaces.h"
#include "bmpx/metadata.h"

#include "plugin_priv.h"

#ifdef HAVE_FTS_H
#  include <fts.h>
#endif

typedef struct {

    GList *local_uris; 
    gchar *uri_original;

} CleanupTuple;

static GList *cleanups = NULL;
#if 0
static guint  guid = -1;
#endif

CleanupTuple*
cleanup_new (const gchar *uri_original)
{
    CleanupTuple *tuple;

    tuple = 
	g_new0 (CleanupTuple,1); 
    tuple->local_uris = NULL;
    tuple->uri_original = 
	g_strdup (uri_original);
    cleanups = g_list_append (cleanups, tuple);
    
    return tuple;
}

void
cleanup_add_uri (CleanupTuple *tuple, const gchar *uri)
{
    tuple->local_uris = g_list_append (tuple->local_uris, g_strdup(uri)); 
}

void
cleanup_clean_up (CleanupTuple *tuple)
{
    for ( ; tuple->local_uris; tuple->local_uris = tuple->local_uris->next)
      {
	gchar *filename;

	filename = g_filename_from_uri (tuple->local_uris->data, NULL, NULL);
	g_unlink (filename);
      }
    g_free (tuple->uri_original);
    g_free (tuple);
}

/* VFS begin */
void
vfs_init(void)
{}

void
vfs_stop(void)
{
    g_list_foreach (cleanups, (GFunc)cleanup_clean_up, NULL);
}

VFSHandle*
vfs_handle_new (const gchar *uri)
{
    VFSHandle *handle;
    GURI *g_uri;

    handle = g_new0 (VFSHandle,1);
    handle->uri_list = NULL;
    handle->uri = g_strdup (uri);
    handle->valid = TRUE;
    handle->need_container = TRUE;

    return handle;
}


static gboolean
find_transport (gpointer key,
                gpointer value,
                gpointer user_data)
{
    VFSHandle *handle = (VFSHandle*)user_data;

    if (bmp_plugin_transport_can_process (value, handle->uri))
	return TRUE;

    return FALSE;
}

static gboolean
find_container (gpointer key,
                gpointer value,
                gpointer user_data)
{
    VFSHandle *handle = (VFSHandle*)user_data;

    if (bmp_plugin_container_can_process (value, handle->uri))
	return TRUE;

    return FALSE;
}
gboolean
vfs_handle_read_uris (VFSHandle *handle)
{
    GObject *plugin_transport = NULL;
    GObject *plugin_container = NULL;
    GError *error = NULL;

    g_return_val_if_fail (handle != NULL, FALSE);
    g_return_val_if_fail (VFS_HANDLE_VALID(handle), FALSE);

    plugin_transport = g_hash_table_find (plugins_table[BMP_PLUGIN_TYPE_TRANSPORT], (GHRFunc) find_transport, handle);
    if (!plugin_transport)
      {
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: No transport found for URI '%s'", G_STRLOC, handle->uri);
	return FALSE;
      }
    else
      {
	if (!bmp_plugin_transport_handle_set  (plugin_transport, handle))
	  {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to set handle (URI: %s) to transport plugin", G_STRLOC, handle->uri);
	    return FALSE;
	  }
	if (!bmp_plugin_transport_handle_read (plugin_transport))
	  {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to process handle (URI: %s) with transport plugin", G_STRLOC, handle->uri);
	    return FALSE;
	  }
      }    

    if (!handle->need_container)
	return TRUE;

    plugin_container = g_hash_table_find (plugins_table[BMP_PLUGIN_TYPE_CONTAINER], (GHRFunc) find_container, handle);
    if (!plugin_container)
      {
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: No container found for URI '%s'", G_STRLOC, handle->uri);
	return FALSE;
      }
    else
      {
	if (!bmp_plugin_container_handle_set  (plugin_container, handle))
	  {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to set handle (URI: %s) to container plugin", G_STRLOC, handle->uri);
	    return FALSE;
	  }

	if (!bmp_plugin_container_handle_read (plugin_container))
	  {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to process handle (URI: %s) with container plugin", G_STRLOC, handle->uri);
	    return FALSE;
	  }
      }    
        
    return TRUE;
}

void
vfs_handle_free (VFSHandle *handle)
{
    g_free (handle->uri);
    g_free (handle->filename);
    g_list_foreach (handle->uri_list, (GFunc)g_free, NULL);
    g_free (handle->buffer);
    g_free (handle);
}

#if 0
/* URI access functions */
VFSFileHandle*
vfs_file_handle_new (const gchar *uri);

#define vfs_file_handle_get_location(handle) (const(handle->handle_path))
#endif
